@@ -1,5 +1,5 @@
 /* wmudmount
- * Copyright © 2010-2014  Brad Jorsch <anomie@users.sourceforge.net>
+ * Copyright © 2010-2018  Brad Jorsch <anomie@users.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -71,7 +71,7 @@ static wmudisks_volume *alloc_volume(UDisksObject *obj) {
     warn(DEBUG_DEBUG, "Allocating new volume %p for %s", v, path);
     v->refct = 0;
     v->path = g_strdup(path);
-    v->obj = g_object_ref(obj);
+    v->obj = UDISKS_OBJECT(g_object_ref(obj));
 
     v->label = NULL;
     v->old_label = NULL;
@@ -106,16 +106,16 @@ static wmudisks_volume *alloc_volume(UDisksObject *obj) {
  */
 static void free_volume_data(wmudisks_volume *v) {
     g_clear_pointer(&v->parent, wmudisks_unref);
-    g_clear_pointer(&v->label, g_free);
-    g_clear_pointer(&v->desc, g_free);
-    g_clear_pointer(&v->uuid, g_free);
-    g_clear_pointer(&v->dev, g_free);
-    g_clear_pointer(&v->alldev, g_strfreev);
+    g_clear_pointer((gpointer*)&v->label, g_free);
+    g_clear_pointer((gpointer*)&v->desc, g_free);
+    g_clear_pointer((gpointer*)&v->uuid, g_free);
+    g_clear_pointer((gpointer*)&v->dev, g_free);
+    g_clear_pointer((gpointer*)&v->alldev, g_strfreev);
     g_clear_pointer(&v->icon, g_object_unref);
     g_clear_pointer(&v->unemblemed_icon, g_object_unref);
-    g_clear_pointer(&v->mountpoint, g_free);
-    g_clear_pointer(&v->allmountpoints, g_strfreev);
-    g_clear_pointer(&v->crypto_device_uuid, g_free);
+    g_clear_pointer((gpointer*)&v->mountpoint, g_free);
+    g_clear_pointer((gpointer*)&v->allmountpoints, g_strfreev);
+    g_clear_pointer((gpointer*)&v->crypto_device_uuid, g_free);
 }
 
 /** Deallocates a volume.
@@ -124,16 +124,15 @@ static void free_volume_data(wmudisks_volume *v) {
 static void dealloc_volume(wmudisks_volume *v) {
     warn(DEBUG_DEBUG, "Deallocating volume %p for %s", v, v->path);
 
-    UDisksObject *obj = UDISKS_OBJECT(v->obj);
-    g_signal_handlers_disconnect_by_func(G_OBJECT(obj), G_CALLBACK(udisks_object_property_changed_handler), v);
+    g_signal_handlers_disconnect_by_func(v->obj, G_CALLBACK(udisks_object_property_changed_handler), v);
     GObject *g;
-    if ((g = G_OBJECT(udisks_object_peek_block(obj)))) {
+    if ((g = G_OBJECT(udisks_object_peek_block(v->obj)))) {
         g_signal_handlers_disconnect_by_func(g, G_CALLBACK(udisks_object_property_changed_handler), v);
     }
-    if ((g = G_OBJECT(udisks_object_peek_filesystem(obj)))) {
+    if ((g = G_OBJECT(udisks_object_peek_filesystem(v->obj)))) {
         g_signal_handlers_disconnect_by_func(g, G_CALLBACK(udisks_object_property_changed_handler), v);
     }
-    if ((g = G_OBJECT(udisks_object_peek_encrypted(obj)))) {
+    if ((g = G_OBJECT(udisks_object_peek_encrypted(v->obj)))) {
         g_signal_handlers_disconnect_by_func(g, G_CALLBACK(udisks_object_property_changed_handler), v);
     }
 
@@ -143,8 +142,8 @@ static void dealloc_volume(wmudisks_volume *v) {
     }
 
     free_volume_data(v);
-    g_clear_pointer(&v->path, g_free);
-    g_clear_pointer(&v->old_label, g_free);
+    g_clear_pointer((gpointer*)&v->path, g_free);
+    g_clear_pointer((gpointer*)&v->old_label, g_free);
     g_clear_pointer(&v->obj, g_object_unref);
     g_slice_free(wmudisks_volume, v);
 }
@@ -154,6 +153,7 @@ static void dealloc_volume(wmudisks_volume *v) {
  */
 static wmudisks_volume *add_volume(UDisksObject *obj) {
     wmudisks_volume *v = alloc_volume(obj);
+    warn(DEBUG_DEBUG, "Adding volume %s (%p)", v->path, v);
     wmudisks_ref(v);
     g_ptr_array_add(all_volumes, v);
     return v;
@@ -164,6 +164,8 @@ static wmudisks_volume *add_volume(UDisksObject *obj) {
  */
 static void remove_volume(wmudisks_volume *v) {
     deactivate_volume(v);
+    warn(DEBUG_DEBUG, "Removing volume %s (%p)", v->path, v);
+    v->removed = 1;
     for (guint i = all_volumes->len; i-- > 0; ) {
         wmudisks_volume *v2 = g_ptr_array_index(all_volumes, i);
         if (v2->parent == v) {
@@ -233,6 +235,7 @@ static gboolean schedule_update_volume_cb(wmudisks_volume *v) {
 }
 static void schedule_update_volume(wmudisks_volume *v) {
     if (!v->update_scheduled) {
+        warn(DEBUG_DEBUG, "Scheduling update for volume %s (%p)", v->path, v);
         v->update_scheduled = g_idle_add_full(G_PRIORITY_HIGH_IDLE,
             (GSourceFunc)schedule_update_volume_cb,
             v, NULL
@@ -240,6 +243,20 @@ static void schedule_update_volume(wmudisks_volume *v) {
     }
 }
 
+/** Wait for an update to the volume
+ * When you expect a change and want to wait for it
+ */
+static void wait_for_update(wmudisks_volume *v) {
+    wmudisks_ref(v);
+    while (!v->update_scheduled && !v->removed) {
+        gtk_main_iteration();
+    }
+    if (v->update_scheduled) {
+        update_volume(v);
+    }
+    wmudisks_unref(v);
+}
+
 /** Get the object for an object-property
  */
 static UDisksObject *object_for_path(const gchar *path) {
@@ -263,6 +280,7 @@ static void update_volume(wmudisks_volume *v) {
     }
 
     warn(DEBUG_DEBUG, " ... Updating volume %s (%p)", v->path, v);
+    if (v->removed) warn(DEBUG_DEBUG, "      !!! Volume was removed!");
 
     // Clear old data
     free_volume_data(v);
@@ -279,7 +297,7 @@ static void update_volume(wmudisks_volume *v) {
     v->can_mount=0;
     v->can_relabel=0;
 
-    UDisksObject *obj = UDISKS_OBJECT(v->obj);
+    UDisksObject *obj = v->obj;
     UDisksBlock *block = udisks_object_peek_block(obj);
     UDisksDrive *drive = udisks_object_peek_drive(obj);
 
@@ -654,7 +672,7 @@ static gboolean wmudisks_unlock_internal(wmudisks_volume *v, wmudisks_callback c
     v->pending_actions++;
 
     udisks_encrypted_call_unlock(
-        udisks_object_peek_encrypted(UDISKS_OBJECT(v->obj)),
+        udisks_object_peek_encrypted(v->obj),
         pass,
         g_variant_new("a{sv}", NULL),
         NULL,
@@ -723,7 +741,7 @@ fail:
         remove_volume(g_ptr_array_index(all_volumes, all_volumes->len - 1));
     }
     if(wmudisks_version){
-        g_clear_pointer(&wmudisks_version, g_free);
+        g_clear_pointer((gpointer*)&wmudisks_version, g_free);
         wmudisks_connection_changed(NULL);
     }
 
@@ -816,13 +834,19 @@ static void wmudisks_mount_cb(UDisksFilesystem *fs, GAsyncResult *res, struct ca
     if (ok) {
         g_clear_error(&err);
         g_set_error(&err, APP_NO_ERROR, 0, "%s mounted on %s", wmudisks_get_display_name(d->req_v), mount_path);
+
+        warn(DEBUG_DEBUG, "Waiting for mount on %s", d->req_v->path);
+        while (!d->req_v->mountpoint) {
+            wait_for_update(d->req_v);
+        }
+        warn(DEBUG_DEBUG, "Got mount on %s", d->req_v->path);
     }
 
     d->req_v->pending_actions--;
     if (d->callback) {
         d->callback(d->req_v, NULL, d->data, !ok, err);
     }
-    g_clear_pointer(&d->req_v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->req_v->old_label, g_free);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
     g_clear_error(&err);
@@ -838,11 +862,19 @@ static void wmudisks_unmount_cb(UDisksFilesystem *fs, GAsyncResult *res, struct
         warn(DEBUG_INFO, "Unmount %s", ok ? "succeeded" : "failed");
     }
 
+    if (ok) {
+        warn(DEBUG_DEBUG, "Waiting for unmount on %s", d->req_v->path);
+        while (!d->req_v->removed && d->req_v->mountpoint) {
+            wait_for_update(d->req_v);
+        }
+        warn(DEBUG_DEBUG, "Got mount on %s", d->req_v->path);
+    }
+
     d->req_v->pending_actions--;
     if (d->callback) {
         d->callback(d->req_v, NULL, d->data, !ok, err);
     }
-    g_clear_pointer(&d->req_v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->req_v->old_label, g_free);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
     g_clear_error(&err);
@@ -858,12 +890,20 @@ static void wmudisks_detach_cb(UDisksDrive *drive, GAsyncResult *res, struct cal
         warn(DEBUG_INFO, "Detach %s", ok ? "succeeded" : "failed");
     }
 
+    if (ok) {
+        warn(DEBUG_DEBUG, "Waiting for detach on %s", d->req_v->path);
+        while (wmudisks_can_detach(d->req_v)) {
+            wait_for_update(d->req_v);
+        }
+        warn(DEBUG_DEBUG, "Got detach on %s", d->req_v->path);
+    }
+
     d->v->pending_actions--;
     d->req_v->pending_actions--;
     if (d->callback) {
         d->callback(d->req_v, d->v, d->data, !ok, err);
     }
-    g_clear_pointer(&d->v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->v->old_label, g_free);
     wmudisks_unref(d->v);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
@@ -880,14 +920,20 @@ static void wmudisks_eject_cb(UDisksDrive *drive, GAsyncResult *res, struct call
         warn(DEBUG_INFO, "Eject %s", ok ? "succeeded" : "failed");
     }
 
-    update_volume(d->v);
+    if (ok) {
+        warn(DEBUG_DEBUG, "Waiting for eject on %s", d->req_v->path);
+        while (wmudisks_can_eject(d->req_v)) {
+            wait_for_update(d->req_v);
+        }
+        warn(DEBUG_DEBUG, "Got eject on %s", d->v->path);
+    }
 
     d->v->pending_actions--;
     d->req_v->pending_actions--;
     if (d->callback) {
         d->callback(d->req_v, d->v, d->data, !ok, err);
     }
-    g_clear_pointer(&d->v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->v->old_label, g_free);
     wmudisks_unref(d->v);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
@@ -904,14 +950,20 @@ static void wmudisks_lock_cb(UDisksEncrypted *enc, GAsyncResult *res, struct cal
         warn(DEBUG_INFO, "Lock %s", ok ? "succeeded" : "failed");
     }
 
-    update_volume(d->v);
+    if (ok) {
+        warn(DEBUG_DEBUG, "Waiting for lock on %s", d->v->path);
+        while (wmudisks_can_lock(d->v)) {
+            wait_for_update(d->v);
+        }
+        warn(DEBUG_DEBUG, "Got lock on %s", d->v->path);
+    }
 
     d->v->pending_actions--;
     d->req_v->pending_actions--;
     if (d->callback) {
         d->callback(d->req_v, d->v, d->data, !ok, err);
     }
-    g_clear_pointer(&d->v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->v->old_label, g_free);
     wmudisks_unref(d->v);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
@@ -973,7 +1025,7 @@ static void wmudisks_relabel_cb(UDisksFilesystem *fs, GAsyncResult *res, struct
     if (d->callback) {
         d->callback(d->req_v, NULL, d->data, !ok, err);
     }
-    g_clear_pointer(&d->req_v->old_label, g_free);
+    g_clear_pointer((gpointer*)&d->req_v->old_label, g_free);
     wmudisks_unref(d->req_v);
     g_slice_free(struct callback_data, d);
     g_clear_error(&err);
@@ -1040,7 +1092,7 @@ gboolean wmudisks_init() {
 }
 
 gboolean wmudisks_mount(wmudisks_volume *v, wmudisks_callback callback, gpointer data) {
-    UDisksBlock *block = udisks_object_peek_block(UDISKS_OBJECT(v->obj));
+    UDisksBlock *block = udisks_object_peek_block(v->obj);
     const gchar *type = udisks_block_get_id_type(block);
     if (!type || !*type) {
         type = "auto";
@@ -1063,7 +1115,7 @@ gboolean wmudisks_mount(wmudisks_volume *v, wmudisks_callback callback, gpointer
     g_variant_builder_add(&b, "{sv}", "fstype", g_variant_new_string(type));
 
     udisks_filesystem_call_mount( 
-        udisks_object_peek_filesystem(UDISKS_OBJECT(v->obj)),
+        udisks_object_peek_filesystem(v->obj),
         g_variant_builder_end(&b),
         NULL,
         (GAsyncReadyCallback)wmudisks_mount_cb,
@@ -1087,7 +1139,7 @@ gboolean wmudisks_unmount(wmudisks_volume *v, wmudisks_callback callback, gpoint
     v->pending_actions++;
 
     udisks_filesystem_call_unmount( 
-        udisks_object_peek_filesystem(UDISKS_OBJECT(v->obj)),
+        udisks_object_peek_filesystem(v->obj),
         g_variant_new("a{sv}", NULL),
         NULL,
         (GAsyncReadyCallback)wmudisks_unmount_cb,
@@ -1100,7 +1152,7 @@ gboolean wmudisks_detach(wmudisks_volume *v, wmudisks_callback callback, gpointe
     wmudisks_volume *req_v = v;
     UDisksDrive *drive = NULL;
     while (TRUE) {
-        drive = udisks_object_peek_drive(UDISKS_OBJECT(v->obj));
+        drive = udisks_object_peek_drive(v->obj);
         if (drive && udisks_drive_get_can_power_off(drive)) {
             break;
         }
@@ -1139,7 +1191,7 @@ gboolean wmudisks_eject(wmudisks_volume *v, wmudisks_callback callback, gpointer
     wmudisks_volume *req_v = v;
     UDisksDrive *drive = NULL;
     while (TRUE) {
-        drive = udisks_object_peek_drive(UDISKS_OBJECT(v->obj));
+        drive = udisks_object_peek_drive(v->obj);
         if (drive && udisks_drive_get_ejectable(drive)) {
             break;
         }
@@ -1178,7 +1230,7 @@ gboolean wmudisks_lock(wmudisks_volume *v, wmudisks_callback callback, gpointer
     wmudisks_volume *req_v = v;
     UDisksEncrypted *enc = NULL;
     while (TRUE) {
-        enc = udisks_object_peek_encrypted(UDISKS_OBJECT(v->obj));
+        enc = udisks_object_peek_encrypted(v->obj);
         if (enc) {
             break;
         }
@@ -1226,7 +1278,7 @@ gboolean wmudisks_unlock(wmudisks_volume *v, wmudisks_callback callback, gpointe
 gboolean wmudisks_forget_passphrase(wmudisks_volume *v, wmudisks_callback callback, gpointer data) {
     wmudisks_volume *req_v = v;
     for (; v; v = v->parent) {
-        if (udisks_object_peek_encrypted(UDISKS_OBJECT(v->obj))) {
+        if (udisks_object_peek_encrypted(v->obj)) {
             warn(DEBUG_INFO, "Forgetting passphrase for %s (on %s)", req_v->path, v->path);
             forget_pass_from_keyring(v);
             if(callback) callback(req_v, v, data, !TRUE, NULL);
@@ -1256,7 +1308,7 @@ gboolean wmudisks_relabel(wmudisks_volume *v, wmudisks_callback callback, gpoint
     v->pending_actions++;
 
     udisks_filesystem_call_set_label( 
-        udisks_object_peek_filesystem(UDISKS_OBJECT(v->obj)),
+        udisks_object_peek_filesystem(v->obj),
         label,
         g_variant_new("a{sv}", NULL),
         NULL,
